open Prolog

let ff = False
let tt = True
let (&&) a b = And (a,b)
let (||) a b = Or (a,b)
let (==) a b = Equals (a,b)

module Basic = struct

  let has_solution = Cps.has_solution prog_empty

  let () =
    assert (has_solution tt) ;
    assert (not (has_solution ff)) ;
    assert (not (has_solution (Atom ("foo",[])))) ;
    assert (has_solution ((ff && ff) || (tt && (ff || tt)))) ;
    Printf.printf "[OK] Basic propositional tests.\n"

  let () =
    let vx = Term.fresh () in
    let vy = Term.fresh () in
    let x = Term.var vx in
    let y = Term.var vy in
    let f t = Term.make "f" [t] in
      assert (has_solution (x == f y)) ;
      assert (has_solution (y == x)) ;
      assert (has_solution ((x == y || x == f y) && (x == f y))) ;
      Printf.printf "[OK] Equality tests.\n"

end

module Test = struct

  let z = Term.make "z" []
  let s x = Term.make "s" [x]

  let lt n m = Atom ("lt",[n;m])
  let plus n m p = Atom ("plus",[n;m;p])
  let ack n m p = Atom ("ack",[n;m;p])
  let even n = Atom ("even",[n])
  let half n m = Atom ("half",[n;m])

  let nil = Term.make "nil" []
  let cons hd tl = Term.make "cons" [hd;tl]

  let (++) = concat_prog
  let (<==) hd body = horn hd body

  let prog =
    prog_empty
    ++
    make_predicate "lt"
      [ forall (fun x ->
          lt z (s x) <== tt) ;
        forall (fun x -> forall (fun y ->
          lt (s x) (s y) <== lt x y)) ]
    ++
    make_predicate "plus"
      [ forall (fun x ->
          plus z x x <== tt) ;
        forall (fun x -> forall (fun y -> forall (fun z ->
          plus (s x) y (s z) <== plus x y z))) ]
    ++
    make_predicate "ack"
      [ forall (fun m ->
          ack z m (s m) <== tt) ;
        forall (fun m -> forall (fun r ->
          ack (s m) z r <== ack m (s z) r)) ;
        forall (fun m -> forall (fun n -> forall (fun r -> forall (fun i ->
          ack (s m) (s n) r <== (ack (s m) n i && ack m i r))))) ]
    ++
    make_predicate "half"
      [ half z z <== tt ;
        forall (fun x -> forall (fun h ->
          half (s (s x)) (s h) <== half x h)) ]
    ++
    make_predicate "even"
      [ even z <== tt ;
        forall (fun x -> even (s (s x)) <== even x) ]

  (** Exécution des tests sur l'interprète donné en argument. *)
  let run has_solution =

    (* Ici on indique au gestionnaire mémoire d'OCaml de ne pas trop
     * travailler, quitte à utiliser plus d'espace. L'objectif est
     * de diminuer le poids du GC dans le temps de calcul des
     * différents interprètes, afin de faire ressortir le gain lié
     * aux changements sur la structure de l'interprète lui même.
     * (Le poids du GC est assez élevé car nous manipulons des très
     * gros termes, et car nous faisons un usage intensif des
     * références.) *)
    Gc.set { (Gc.get ()) with
               Gc.minor_heap_size = 2048000 ;
               Gc.space_overhead = max_int } ;

    let has_solution = has_solution prog in

    let rec of_nat = function
      | 0 -> z
      | n -> s (of_nat (n-1))
    in

    let x = Term.var (Term.fresh ()) in
    assert (has_solution (lt (of_nat 2) (of_nat 3))) ;
    assert (has_solution (plus (of_nat 2) (of_nat 3) (of_nat 5))) ;
    assert (has_solution (plus (of_nat 2) (of_nat 3) x)) ;
    assert (not (has_solution ((plus (of_nat 2) (of_nat 3) x) &&
                               (lt x (of_nat 5) || lt (of_nat 5) x)))) ;
    Printf.printf "[OK] Basic Prolog\n" ;

    let x = Term.var (Term.fresh ()) in
    let h = Term.var (Term.fresh ()) in
    let t0 = Unix.gettimeofday () in
    assert (not (has_solution (ack (of_nat 2) (of_nat 2) (s x) &&
                               half x h && even h))) ;
    assert (has_solution (ack (of_nat 2) (of_nat 3) (s x) &&
                          half x h && even h)) ;
    assert (has_solution (ack (of_nat 3) (of_nat 5) x)) ;
    assert (not (has_solution (ack (of_nat 3) (of_nat 5) z))) ;
    Printf.printf "[OK] Ackermann (%.2fs)\n" (Unix.gettimeofday () -. t0) ;

    let x = Term.var (Term.fresh ()) in
    let y = Term.var (Term.fresh ()) in
    let t0 = Unix.gettimeofday () in
    assert (not (has_solution (plus x y (of_nat 66666) &&
                               x == z && y == z))) ;
    Printf.printf "[OK] 0+0=... (%.2fs)\n" (Unix.gettimeofday () -. t0)

end

let () =
  Printf.printf "Testing CPS implementation...\n%!" ;
  Test.run Cps.has_solution

(* Décommenter pour tester dans la partie 5...

let () =
  Printf.printf "Testing defunctionalized implementation...\n%!" ;
  Test.run Defunc.has_solution

let () =
  Printf.printf "Testing abstract machine...\n%!" ;
  Test.run Abstract_machine.has_solution

 *)
